#!/usr/bin/python

import sys
import copy
import random
import gc

fff = open(sys.argv[1], "r")

random.seed()
gc.enable()


class Spec:
    def __init__(self, v):
        self.v = v

class Bit(Spec):
    def __init__(self, src, pos):
        self.src = src
        self.pos = pos
        self.origin = None

    def __str__(self):
        if self.pos != -1:
            return "%s [%d]" % (self.src,self.pos)#,str(self.origin))
        else:
            return "%s" % self.src

class SpecOp(Spec):
    def __init__(self, v):
        self.v = v
        
class Not(SpecOp):
    def __init__(self, a):
        self.a = a

    def __str__(self):
        return "NOT(" + str(self.a) + ")"
        
class And(SpecOp):
    def __init__(self, l, r):
        self.l = l
        self.r = r
    
    def __str__(self):
        return "AND("+str(self.l)+","+str(self.r)+")"

class Xor(SpecOp):
    def __init__(self, l, r):
        self.l = l
        self.r = r

    def __str__(self):
        return "XOR("+str(self.l)+","+str(self.r)+")"

class Or(SpecOp):
    def __init__(self, l, r):
        self.l = l
        self.r = r

    def __str__(self):
        return "OR("+str(self.l)+","+str(self.r)+")"

class Sum(SpecOp):
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def __str__(self):
        return "SUM("+str(self.a)+","+str(self.b)+","+str(self.c)+")"

class Reg:
    def __init__(self, size, imm=None, src=None):
        self.size = size
        self.content = [0]*self.size
        if imm is not None:
            self.set_content(imm)
        if src is not None:
            self.set_special_content(src)
        self.origin = None

    def __str__(self):
        return ";".join(str(k) for k in self.content)

    def print_byte(self, n):
        return ",".join(str(k) for k in self.content[n:n+8])

    def set_content(self, imm):
        i=0
        while imm > 0:
            self.content[i] = imm&1
            imm = imm>>1
            i += 1

    def check_add(self, a, b):
        
        if isinstance(a, And) or isinstance(a, Xor):
            if isinstance(b, And) or isinstance(b, Xor):
                if not (isinstance(a.l, SpecOp) or isinstance(a.r, SpecOp)) and ( (isinstance(b.r, Xor) or isinstance(b.r, And)) or (isinstance(b.l, Xor) or isinstance(b.l, And)) ):
                    if not (isinstance(a.l, Spec) and isinstance(a.r, Spec)):
                        return None
                    s1 = a.l.src
                    p1 = a.l.pos
                    s2 = a.r.src
                    p2 = a.r.pos
                    if (isinstance(b.l, And) or isinstance(b.l, Xor)) and not (isinstance(b.l.l, SpecOp) or isinstance(b.l.r, SpecOp)):
                        if b.l.l.src == s1 and b.l.l.pos == p1 and b.l.r.src == s2 and b.l.r.pos == p2:
                            return Sum(a.l, a.r, b.r)
                    if (isinstance(b.r, And) or isinstance(b.r, Xor)) and not (isinstance(b.r.l, SpecOp) or isinstance(b.r.r, SpecOp)):
                        if b.r.l.src == s1 and b.r.l.pos == p1 and b.r.r.src == s2 and b.r.r.pos == p2:
                            return Sum(a.l, a.r, b.l)
                elif not (isinstance(b.l, SpecOp) or isinstance(b.r, SpecOp) ) and ( (isinstance(a.r, Xor) or isinstance(a.r, And)) or (isinstance(a.l, Xor) or isinstance(a.l, And)) ):
                    if not (isinstance(b.l, Spec) and isinstance(b.r, Spec)):
                        return None
                    s1 = b.l.src
                    p1 = b.l.pos
                    s2 = b.r.src
                    p2 = b.r.pos
                    if (isinstance(a.l, And) or isinstance(a.l, Xor)) and not (isinstance(a.l.l, SpecOp) or isinstance(a.l.r, SpecOp)):
                        if a.l.l.src == s1 and a.l.l.pos == p1 and a.l.r.src == s2 and a.l.r.pos == p2:
                            return Sum(b.l, b.r, a.r)
                    if (isinstance(a.r, And) or isinstance(a.r, Xor)) and not (isinstance(a.r.l, SpecOp) or isinstance(a.r.r, SpecOp)):
                        if a.r.l.src == s1 and a.r.l.pos == p1 and a.r.r.src == s2 and a.r.r.pos == p2:
                            return Sum(b.l, b.r, a.l)

        return None
        
    def set_special_content(self, src):
        for i in xrange(self.size):
            self.content[i] = Bit(src, i)

    def zero_content(self):
        for k in xrange(self.size):
            if isinstance(self.content[k], Spec):
                a = self.content[k]
                self.content[k] = 0
                del a
            self.content[k] = 0
            
    def pinsrw(self, reg, n):
        ###SIMP
        self.zero_content()
        ###
        for i in xrange(16):
            self.content[i+n*16] = reg.content[i]

    def pextrw(self, reg, n):
        ###SIMP
        self.zero_content()
        ###
        for i in xrange(16):
            self.content[i] = reg.content[i+n*16]

    def get_byte(self, n):
        bval = 0
        off = 0
        try:
            for k in xrange(n,n+8,1):
                bval += (self.content[k]<<off)
                off += 1
            return bval
        except:
            return None
            
    def pshufb(self, reg):
        temp = copy.copy(self.content)
        for k in xrange(16):
            off = reg.get_byte(k*8)
            if off is None:
                off = 0xff
            if (off&0x80) != 1:
                off = off&0xf
                for i in xrange(8):
                    self.content[k*8 + i] = temp[off*8 + i]
            else:
                for i in xrange(8):
                    self.content[k*8 + i] = 0
        del temp

    def pand(self, reg):
        for k in xrange(self.size):
            v = reg.content[k]
            if isinstance(v, Spec) and isinstance(self.content[k], Spec):
                x = self.check_add(self.content[k],v)
                if x is None:
                    self.content[k] = And(self.content[k], v)
                else:
                    self.content[k] = x
            elif isinstance(v, Spec):
                self.content[k] = v if self.content[k] == 1 else 0
            elif isinstance(self.content[k], Spec):
                if v==0:
                    self.content[k] = 0
            else:
                self.content[k] &= v

    def por(self, reg):
        for k in xrange(self.size):
            v = reg.content[k]
            if isinstance(v, Spec) and isinstance(self.content[k], Spec):
                self.content[k] = Or(self.content[k], v)
            elif isinstance(v, Spec):
                self.content[k] = 1 if self.content[k] == 1 else v
            elif isinstance(self.content[k], Spec):
                if v==1:
                    self.content[k] = 1
            else:
                self.content[k] |= v 

    def pxor(self, reg):
        for k in xrange(self.size):
            v = reg.content[k]
            if not isinstance(v, Spec) and not isinstance(self.content[k], Spec):
                self.content[k] ^= v
            elif not isinstance(v, Spec):
                if v == 0:
                    self.content[k] = self.content[k]
                else:
                    if isinstance(self.content[k], Not):
                        self.content[k] = self.content[k].a
                    else:
                        self.content[k] = Not(self.content[k])
            elif not isinstance(self.content[k], Spec):
                if self.content[k] == 0:
                    self.content[k] = reg.content[k]
                else:
                    if isinstance(reg.content[k], Not):
                        self.content[k] = reg.content[k].a
                    else:
                        self.content[k] = Not(reg.content[k])
            else:
                if isinstance(self.content[k], Not) and isinstance(v, Not):
                    a = self.content[k].a
                    b = v.a
                else:
                    a = self.content[k]
                    b = v

                if ( (isinstance(a,And) and isinstance(b,Xor)) or (isinstance(b,And) and isinstance(a,Xor)) ) and (a.l == b.l and a.r == b.r):
                    self.content[k] = Or(a.l, a.r)
                    return
                    
                x = self.check_add(a,b)
                if x is None:
                    self.content[k] = Xor(a, b)
                else:
                    self.content[k] = x

    def paddb(self, reg):
        temp = copy.copy(self.content)
        self.zero_content()
        for k in xrange(self.size/8):
            a = self.get_byte(k*8)
            b = reg.get_byte(k*8)
            if a is not None and b is not None:
                x = (a+b)&0xff
                for i in xrange(8):
                    self.content[k*8 + i] = (x>>i)&1
            elif a is not None and a == 0:
                for i in xrange(8):
                    self.content[k*8 + i] = reg.content[k*8 + i]
            elif b is not None and b == 0:
                for i in xrange(8):
                    self.content[k*8 + i] = temp[k*8 + i]
            else:
                for i in xrange(8):
                    self.content[k*8 + i] = Bit("ADDB("+str(temp[k*8 + i])+","+str(reg.content[k*8 + i])+")", -1)
        del temp
                    
    def pandn(self, reg):
        self.non()
        self.pand(reg)

    def psllq(self, n):
        temp = copy.copy(self.content)
        self.content = [0]*self.size
        for k in xrange(2):
            for i in xrange(n, 64):
                self.content[k*64 + i] = temp[k*64 + i - n]
        del temp

    def psrlq(self, n):
        temp = copy.copy(self.content)
        self.content = [0]*self.size
        for k in xrange(2):
            for i in xrange(0, 64-n):
                self.content[k*64 + i] = temp[k*64 + i + n]
        del temp

    def pminub(self, reg):
        temp = copy.copy(self.content)
        for k in xrange(16):
            a = self.get_byte(k*8)
            b = reg.get_byte(k*8)
            if (a is not None and a == 0) or (b is not None and b == 0xff) :
                for i in xrange(8):
                    self.content[k*8 + i] = self.content[k*8 + i]
            elif (b is not None and b == 0) or (a is not None and a == 0xff):
                for i in xrange(8):
                    self.content[k*8 + i] = reg.content[k*8 + i]
            elif a is not None and b is not None:
                x = a if a<b else b
                for i in xrange(8):
                    self.content[k*8 + i] = (x&1)
                    x = x>>1
            else:
                #YOLO
                for i in xrange(8):
                    self.content[k*8+i] = Bit("PMINUB("+str(self.content[k*8+i])+","+str(reg.content[k*8+i])+")", -1)
        del temp

    def pmaxub(self, reg):
        temp = copy.copy(self.content)
        for k in xrange(16):
            a = self.get_byte(k*8)
            b = reg.get_byte(k*8)
            if (a is not None and a == 0) or (b is not None and b == 0xff) :
                for i in xrange(8):
                    self.content[k*8 + i] = reg.content[k*8 + i]
            elif (b is not None and b == 0) or (a is not None and a == 0xff):
                for i in xrange(8):
                    self.content[k*8 + i] = self.content[k*8 + i]
            elif a is not None and b is not None:
                x = a if a>b else b
                for i in xrange(8):
                    self.content[k*8 + i] = (x&1)
                    x = x>>1
            else:
                #YOLO
                for i in xrange(8):
                    self.content[k*8+i] = Bit("PMAXUB("+str(self.content[k*8+i])+","+str(reg.content[k*8+i])+")", -1)
        del temp

    def rol(self, n):
        temp = copy.copy(self.content)
        self.content = [0]*self.size
        for k in xrange(self.size):
            self.content[k] = temp[(k-n)%self.size]
        del temp

    def ror(self, n):
        temp = copy.copy(self.content)
        self.content = [0]*self.size
        for k in xrange(self.size):
            self.content[k] = temp[(k+n)%self.size]
        del temp

    def non(self):
        for k in xrange(self.size):
            if isinstance(self.content[k], Spec):
                if isinstance(self.content[k], Not):
                    self.content[k] = self.content[k].a
                else:
                    self.content[k] = Not(self.content[k])
            else:
                self.content[k] ^= 1

    def copy(self, reg):
        self.content = [0]*self.size
        for k in xrange(min(self.size, reg.size)):
            self.content[k] = reg.content[k]

    def copyfromup(self, reg):
        self.content = [0]*self.size
        for k in xrange(self.size):
            self.content[k] = reg.content[self.size+k]

    def copytoup(self, reg):
        for k in xrange(reg.size):
            self.content[reg.size+k] = reg.content[k]

regs = {}
stack_ind = 0
stack_switch = 0
stack = [{},{}]

special_regs = []
for k in xrange(8):
    special_regs.append("xmm%d" % (k+8))
    special_regs.append("r%d" % (k+8))

def prettyprint(x):
    for k in x:
        print "\t%s"%k,x[k]


def getreg(x):
    if x[0] == "r":
        return x
    elif x[0] == "e":
        return "r"+x[1:]
    elif x.startswith("xmm"):
        return x
    else:
        return "r"+x

def reg_size(x):
    if x.startswith("xmm"):
        return 128
    return 64



if len(sys.argv) <= 2:
    for k in xrange(8,16,1):
        regs["r%d"%k] = Reg(64)
    for k in ("ax","bx","cx","dx","di","si"):
        regs["r%s"%k] = Reg(64) 
    for k in xrange(16):
        regs["xmm%d"%k] = Reg(128)
else:

    for k in xrange(16):
        regs["xmm%d"%k] = Reg(128, src="xmm%d"%k)
    for k in xrange(8,16,1):
        regs["r%d"%k] = Reg(64,src="r%d"%k)
    for k in ("ax","bx","cx","dx","di","si"):
        regs["r%s"%k] = Reg(64, src= "r%s"%k)

def isimm(x):
    return x.startswith("0x")

def isder(x):
    return x.startswith("[")

def print_xmm():
    for k in xrange(8):
        print "ARG xmm%d %s" %( k+8, regs["xmm%d"%(k+8)])

def reset_xmm():
    for k in regs:
        if k.startswith("xmm"):
            for i in xrange(regs[k].size):
                if isinstance(regs[k].content[i], Spec):
                    regs[k].content[i] = Bit(k, i)

for ll in fff:
 try:
    ll =ll.rstrip('\n')
    l=ll.split(" ")
    try:
        while l.index('') != 213123:
            l.remove('')
    except:
        pass
    if len(l) == 0:
        continue
    ins=l[0]
    if ins in ("push","sub","pop"):
        continue
    elif ins in ("jmp","jne","jnz","lea","xchg"):
        continue
    elif ins == "int":
        stack_ind ^= 1
        stack_switch += 1
        continue
    ind = 1
    try:
        while not "," in l[ind]:
            ind += 1
        ops = l[ind].split(",")
    except:
        ops = [l[ind-1],]
    if len(ops) == 2 and ind != len(l)-1:
        ops = [ops[0],l[-1]]
    print "INS:",ins,ops

    if ins.startswith("mov"):
        if ops[0] == 'rbp':
            continue
        if not isimm(ops[1]) and not isder(ops[1]):
            if not isimm(ops[0]) and not isder(ops[0]):
                regs[getreg(ops[0])].copy(regs[getreg(ops[1])])
                if getreg(ops[0]) == "r15":
                    print_xmm()
            elif isder(ops[0]) and ("rsp" in ops[0] or "rbp" in ops[0]):
                print "WRITE",regs[getreg(ops[1])]
                if "+" in ops[0]:
                    off = int(ops[0].split("+")[1][:-1], 16)
                else:
                    off = 0
                stack[stack_ind][off] = Reg(64)
                stack[stack_ind][off].copy(regs[getreg(ops[1])])
                if regs[getreg(ops[1])].size > 64:
                    stack[stack_ind][off+8] = Reg(64)
                    stack[stack_ind][off+8].copyfromup(regs[getreg(ops[1])])
            else:
                print "unhandled case"
                sys.exit(1)
        elif isimm(ops[1]):
            if not isimm(ops[0]) and not isder(ops[0]):
                regs[getreg(ops[0])] = Reg(reg_size(getreg(ops[0])), imm=int(ops[1],16))
            else:
                print "unhandled case"
                sys.exit(1)
        else:
            if not isimm(ops[0]) and not isder(ops[0]):
                if 'rbpXXXnotusedanymoreXXX' in ops[1]:
                    if "+" in ops[1]:
                        off = int(ops[1].split("+")[1][:-1], 16)
                    else:
                        off = 0
                    off -= 0x10
                    regs[getreg(ops[0])].copy(stack[stack_ind^1][off])
                    if regs[getreg(ops[0])].size > 64:
                        regs[getreg(ops[0])].copytoup(stack[stack_ind^1][off+8])
                    #print "READ", ops[1], regs[getreg(ops[0])]
                else:        
                    regs[getreg(ops[0])] = Reg(reg_size(getreg(ops[0])), src=ops[1])#+"(%d)"%stack_switch)
            else:
                print "unhandled case"
                sys.exit(1)
    elif "xor" in ins:
        if ops[0] == ops[1]:
            regs[getreg(ops[0])] = Reg(reg_size(getreg(ops[0])))
        else:
            regs[getreg(ops[0])].pxor(regs[getreg(ops[1])])
    elif ins == "rol":
        d=int(ops[1],16)
        regs[getreg(ops[0])].rol(d)
    elif ins == "ror":
        d=int(ops[1],16)
        regs[getreg(ops[0])].ror(d)
    elif ins == "not":
        regs[getreg(ops[0])].non()

    elif ins == "addXXXnotusedanymoreXXX":
        if "rsp" in ops[0]:
            continue
        else:
           print "unhandled case x"
           sys.exit(1) 
        
    elif ins == "pinsrw":
        d=int(ops[2],16)
        regs[getreg(ops[0])].pinsrw(regs[getreg(ops[1])], d)
    elif ins == "pextrw":
        d=int(ops[2],16)
        regs[getreg(ops[0])].pextrw(regs[getreg(ops[1])], d)
    elif ins == "pshufb":
        regs[getreg(ops[0])].pshufb(regs[getreg(ops[1])])
    elif ins == "pandn":
        regs[getreg(ops[0])].pandn(regs[getreg(ops[1])])
    elif ins == "paddb" or ins == "add":
        regs[getreg(ops[0])].paddb(regs[getreg(ops[1])])
    elif ins == "pand" or ins == "and":
        if isder(ops[0]) and "rbp" in ops[0]:
            if "-" in ops[1]:
                off = 0 - int(ops[1].split("-")[1][:-1], 16)
            else:
                off = 0
            off += 0x10
            x=Reg(64)
            x.copy(stack[stack_ind][off])
            x.pand(regs[getreg(ops[1])])
            stack[stack_ind][off] = x
        else:
            regs[getreg(ops[0])].pand(regs[getreg(ops[1])])
    elif ins == "por" or ins == "or":
        if isder(ops[0]) and "rbp" in ops[0]:
            if "-" in ops[1]:
                off = 0 - int(ops[1].split("-")[1][:-1], 16)
            else:
                off = 0
            off += 0x10
            x=Reg(64)
            x.copy(stack[stack_ind][off])
            x.por(regs[getreg(ops[1])])
            stack[stack_ind][off] = x
        else:
            regs[getreg(ops[0])].por(regs[getreg(ops[1])])
    elif ins == "pminub":
        regs[getreg(ops[0])].pminub(regs[getreg(ops[1])])
    elif ins == "pmaxub":
        regs[getreg(ops[0])].pmaxub(regs[getreg(ops[1])])
    elif ins == "psllq":
        d=int(ops[1],16)
        regs[getreg(ops[0])].psllq(d)
    elif ins == "psrlq":
        d=int(ops[1],16)
        regs[getreg(ops[0])].psrlq(d)
    elif ins == "pcmpeqb":
        if ops[0] == ops[1]:
            regs[getreg(ops[0])] = Reg(128, imm=0xffffffffffffffffffffffffffffffff)
        else:
            print "unhandled case x"
            sys.exit(1)
    elif ins == "test":
        r0 = Reg(regs[getreg(ops[0])].size)
        r0.copy(regs[getreg(ops[0])])
        if isimm(ops[1]):
            r1 = Reg(regs[getreg(ops[0])].size, imm=int(ops[1],16))
            r0.pand(r1)
        else:
            r0.pand(regs[getreg(ops[1])])
        print "TESTED",getreg(ops[0]),r0
    else:
        print "unhandled case"
        sys.exit(1)

    #prettyprint(regs)
    gc.collect()
 except:
     print "Error :)"
     prettyprint(regs)
